# -*- coding: utf-8 -*-

import requests
from . import errors
from . import xmlparse
from . import response


class Client(object):
    def __init__(self, url=None, maximum_records=10, record_schema=None):
        self.url = url
        self.maximum_records = maximum_records
        self.sru_version = '1.2'
        self.record_schema = record_schema

    def searchretrieve(self, query, start_record=1, requests_kwargs=None):
        params = {
            'operation': 'searchRetrieve',
            'version': self.sru_version,
            'query': query,
            'startRecord': start_record,
            'maximumRecords': self.maximum_records,
        }

        if self.record_schema:
            params['recordSchema'] = self.record_schema

        data_loader = DataLoader(self.url, params, requests_kwargs)
        return response.SearchRetrieveResponse(data_loader)

    def explain(self, requests_kwargs=None):
        params = {
            'operation': 'explain',
            'version': self.sru_version,
        }
        data_loader = DataLoader(self.url, params, requests_kwargs)
        return response.ExplainResponse(data_loader)


class DataLoader(object):
    def __init__(self, url, params, requests_kwargs=None):
        self.session = requests.Session()
        self.url = url
        self.params = params
        self.response = None
        self.xmlparser = xmlparse.XMLParser()
        self.requests_kwargs = requests_kwargs or {}

    def load(self, **kwargs):
        self.params.update(kwargs)
        xml = self._get_content(self.url, self.params)
        self._check_errors(xml)
        return xml

    def _get_content(self, url, params):
        try:
            res = self.session.get(
                url,
                params=params,
                **self.requests_kwargs
            )
            res.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise errors.SruthiError("HTTP error: %s" % e)
        except requests.exceptions.RequestException as e:
            raise errors.SruthiError("Request error: %s" % e)

        return self.xmlparser.parse(res.content)

    def _check_errors(self, xml):
        sru = '{http://www.loc.gov/zing/srw/}'
        diag = '{http://www.loc.gov/zing/srw/diagnostic/}'
        diagnostics = self.xmlparser.find(
            xml,
            f'{sru}diagnostics/{diag}diagnostic'
        )
        if diagnostics:
            error_msg = ", ".join([d.text for d in diagnostics])
            raise errors.SruError(error_msg)



__version__ = '0.1.2'
__all__ = ['client', 'errors', 'response', 'sru', 'xmlparse']

from .errors import SruthiError, ServerIncompatibleError, SruError, NoMoreRecordsError  # noqa
from .errors import SruthiWarning, WrongNamespaceWarning # noqa
from .sru import searchretrieve, explain  # noqa
from .client import Client # noqa



import re
from xml.etree.ElementTree import Element
import defusedxml.ElementTree as etree
import xmltodict
from . import errors


class XMLNone(object):
    def __nonzero__(self):
        return False

    def __bool__(self):
        return False

    def iter(self):
        return []

    text = None


class XMLParser(object):
    def __init__(self):
        self.namespaces = {
            'sru': 'http://www.loc.gov/zing/srw/',
            'isad': 'http://www.expertisecentrumdavid.be/xmlschemas/isad.xsd',
            'rel': 'info:srw/extension/2/relevancy-1.0',
            'ap': 'http://www.archivportal.ch/srw/extension/',
            'zr': 'http://explain.z3950.org/dtd/2.1/',
        }
        self.dict_namespaces = {
            'http://www.loc.gov/zing/srw/': 'sru',
            'http://explain.z3950.org/dtd/2.1/': 'zr',
            'info:srw/extension/2/relevancy-1.0': None,
            'http://www.archivportal.ch/srw/extension/': None,
            'http://www.loc.gov/MARC21/slim': None,
            'http://www.loc.gov/mods/v3': None,
            'http://www.loc.gov/standards/mods/v3/mods-3-6.xsd': None,
            'http://www.loc.gov/standards/mods/v3/mods-3-6.xsd': None,
            'http://purl.org/dc/elements/1.1/': None,
            'http://www.expertisecentrumdavid.be/xmlschemas/isad.xsd': None,
            'http://www.w3.org/2001/XMLSchema-instance': None,
            'http://www.w3.org/XML/1998/namespace': None,
        }

    def parse(self, content):
        try:
            return etree.fromstring(content)
        except Exception as e:
            raise errors.SruError("Error while parsing XML: %s" % e)

    def find(self, xml, path):
        if isinstance(path, list):
            for p in path:
                elem = self.find(xml, p)
                if not isinstance(elem, XMLNone):
                    return elem
            return XMLNone()
        elem = xml.find(path, self.namespaces)
        if elem is None:
            return XMLNone()
        return elem

    def findall(self, xml, path):
        return xml.findall(path, self.namespaces)

    def tostring(self, xml):
        return etree.tostring(xml)

    def todict(self, xml, **kwargs):
        if isinstance(xml, XMLNone):
            return None
        if isinstance(xml, Element):
            xml = self.tostring(xml)

        dict_args = {
            'dict_constructor': dict,
            'process_namespaces': True,
            'namespaces': self.dict_namespaces,
            'attr_prefix': '',
            'cdata_key': 'text',
        }
        dict_args.update(kwargs)
        return dict(xmltodict.parse(xml, **dict_args))

    def namespace(self, element):
        m = re.match(r'\{(.*)\}', element.tag)
        return m.group(1) if m else ''



# -*- coding: utf-8 -*-

from collections import defaultdict
import re
import warnings
from flatten_dict import flatten
from . import xmlparse
from . import errors


class Response(object):
    def __init__(self, data_loader):
        self.data_loader = data_loader
        self.xmlparser = xmlparse.XMLParser()
        self.records = []
        xml = self.data_loader.load()
        self._parse_content(xml)

    def maybe_int(self, s):
        try:
            return int(s)
        except (ValueError, TypeError):
            return s

    def _check_response_tag(self, xml, tag):
        sru = '{http://www.loc.gov/zing/srw/}'
        response = f"{sru}{tag}"
        if not xml.tag == response:
            # fix namespace for servers that provide the wrong namespace URI
            main_ns = self.xmlparser.namespace(xml)
            if 'www.loc.gov/zing/srw' in main_ns:
                warnings.warn(
                    f"""
                    The server has the wrong namespace for SRU,
                    it should be {sru} but it's currently set to {{{main_ns}}}.
                    """,
                    errors.WrongNamespaceWarning
                )
                self.xmlparser.namespaces['sru'] = main_ns
            else:
                raise errors.ServerIncompatibleError(
                    f"Server response did not contain a {response} tag"
                )


class SearchRetrieveResponse(Response):
    def __repr__(self):
        try:
            return (
                'SearchRetrieveResponse('
                'sru_version=%r,'
                'count=%r,'
                'next_start_record=%r)'
                ) % (
                   self.sru_version,
                   self.count,
                   self.next_start_record,
                )
        except AttributeError:
            return 'SearchRetrieveResponse(empty)'

    def _parse_content(self, xml):
        self._check_response_tag(xml, 'searchRetrieveResponse')

        self.sru_version = self.xmlparser.find(xml, './sru:version').text
        self.count = self.maybe_int(self.xmlparser.find(xml, './sru:numberOfRecords').text)
        self._extract_records(xml)

        next_start_record = self.xmlparser.find(xml, './sru:nextRecordPosition').text
        if next_start_record:
            self.next_start_record = self.maybe_int(next_start_record)
        else:
            self.next_start_record = None

    def __length_hint__(self):
        return self.count

    def __iter__(self):
        # use while loop since self.records could grow while iterating
        i = 0
        while True:
            # load new data when near end
            if i == len(self.records):
                try:
                    self._load_new_data()
                except errors.NoMoreRecordsError:
                    break
            yield self.records[i]
            i += 1

    def __getitem__(self, key):
        if isinstance(key, slice):
            limit = max(key.start or 0, key.stop or self.count)
            self._load_new_data_until(limit)
            count = len(self.records)
            return [self.records[k] for k in range(*key.indices(count))]

        if not isinstance(key, int):
            raise TypeError("Index must be an integer or slice")

        limit = key
        if limit < 0:
            # if we get a negative index, load all data
            limit = self.count
        self._load_new_data_until(limit)
        return self.records[key]

    def _load_new_data_until(self, limit):
        while limit >= len(self.records):
            try:
                self._load_new_data()
            except errors.NoMoreRecordsError:
                break

    def _load_new_data(self):
        if self.next_start_record is None:
            raise errors.NoMoreRecordsError()
        xml = self.data_loader.load(startRecord=self.next_start_record)
        self._parse_content(xml)

    def _extract_records(self, xml):
        new_records = []

        xml_recs = self.xmlparser.findall(xml, './sru:records/sru:record')
        for xml_rec in xml_recs:
            record = defaultdict()
            record['schema'] = self.xmlparser.find(xml_rec, './sru:recordSchema').text
            record_data = self.xmlparser.find(xml_rec, './sru:recordData')
            extra_data = self.xmlparser.find(xml_rec, './sru:extraRecordData')

            record.update(self._tag_data(record_data, 'sru:recordData') or {})
            record['extra'] = self._tag_data(extra_data, 'sru:extraRecordData')

            record = dict(record)
            new_records.append(record)
        self.records.extend(new_records)

    def _tag_data(self, elem, parent):
        if not elem:
            return None

        record_data = self.xmlparser.todict(elem, xml_attribs=True).get(parent)
        if not record_data:
            return None

        # check if there is only one element on the top level
        keys = list(record_data.keys())
        if len(record_data) == 1 and len(keys) > 0 and len(record_data[keys[0]]) > 0:
            record_data = record_data[keys[0]]

        record_data.pop('schemaLocation', None)
        record_data.pop('xmlns', None)

        def leaf_reducer(k1, k2):
            # only use key of leaf element
            return k2

        try:
            record_data = flatten(record_data, reducer=leaf_reducer)
        except ValueError:
            # if the keys of the leaf elements are not unique
            # the dict will not be flattened
            pass

        return record_data

    def _remove_namespace(self, elem):
        ns_pattern = re.compile('{.+}')
        tag_name = ns_pattern.sub('', elem.tag)
        return tag_name


class ExplainResponse(Response):
    def __repr__(self):
        return (
            'ExplainResponse('
            'sru_version=%r,'
            'server=%r,'
            'database=%r'
            'index=%r'
            'schema=%r'
            'config=%r)'
            ) % (
               self.sru_version,
               self.server,
               self.database,
               self.index,
               self.schema,
               self.config,
            )

    def _parse_content(self, xml):
        self._check_response_tag(xml, 'explainResponse')

        record_schema = self.xmlparser.find(xml, './/sru:recordSchema').text
        if record_schema:
            self.xmlparser.namespaces['zr'] = record_schema

        self.sru_version = self.xmlparser.find(xml, './sru:version').text

        self.server = self._parse_server(xml)
        self.database = self._parse_database(xml)
        self.index = self._parse_index(xml)
        self.schema = self._parse_schema(xml)
        self.config = self._parse_config(xml)

    def _parse_server(self, xml):
        server_info = {
            'host': self.xmlparser.find(xml, './/zr:serverInfo/zr:host').text,
            'port': self.xmlparser.find(xml, './/zr:serverInfo/zr:port').text,
        }
        server_info['port'] = self.maybe_int(server_info['port'])
        return server_info

    def _parse_schema(self, xml):
        def bool_or_none(v):
            if v is None:
                return None
            return bool(v)

        def ident(a):
            return a

        attributes = {
            'identifier': ident,
            'name': ident,
            'location': ident,
            'sort': bool_or_none,
            'retrieve': bool_or_none,
        }

        schemas = {}
        for schema in self.xmlparser.findall(xml, './/zr:schemaInfo/zr:schema'):
            schema_info = {}
            for attr, fn in attributes.items():
                xml_attr = schema.attrib.get(attr)
                if xml_attr:
                    schema_info[attr] = fn(xml_attr)
            schema_info['title'] = self.xmlparser.find(schema, './zr:title').text
            schemas[schema.attrib.get('name')] = schema_info
        return schemas

    def _parse_config(self, xml):
        config = {}
        for setting in self.xmlparser.findall(xml, './/zr:configInfo/zr:setting'):
            t = setting.attrib['type']
            config[t] = self.maybe_int(setting.text)

        # defaults
        defaults = {}
        for default in self.xmlparser.findall(xml, './/zr:configInfo/zr:default'):
            t = default.attrib['type']
            defaults[t] = self.maybe_int(default.text)
        config['defaults'] = defaults
        return config

    def _parse_database(self, xml):
        db = self.xmlparser.find(xml, './/zr:databaseInfo')
        db_info = {
            'title': self.xmlparser.find(db, ['./zr:title', './title']).text,
            'description': self.xmlparser.find(db, ['./zr:description', './description']).text,
            'contact': self.xmlparser.find(db, ['./zr:contact', './contact']).text,
        }
        db_info = {k: v.strip() if v else v for (k, v) in db_info.items()}
        return db_info

    def _parse_index(self, xml):
        index = defaultdict(defaultdict)
        for index_set in self.xmlparser.findall(xml, './/zr:indexInfo/zr:set'):
            index[index_set.attrib['name']] = defaultdict()

        for index_field in self.xmlparser.findall(xml, './/zr:indexInfo/zr:index'):
            title = self.xmlparser.find(index_field, './zr:title').text or \
                    self.xmlparser.find(index_field, './title').text
            if title:
                title = title.strip()
            for name in self.xmlparser.findall(index_field, './/zr:map/zr:name'):
                index[name.attrib['set']][name.text.strip()] = title

        return {k: dict(v) for k, v in dict(index).items()}



class SruthiError(Exception):
    """
    General sruthi error class to provide a superclass for all other errors
    """


class ServerIncompatibleError(SruthiError):
    """
    The error raised from sru.search/sru.explain when the server doesn't behave
    like a SRU endpoint.
    """


class SruError(SruthiError):
    """
    The error raised from sru.search/sru.explain when the SRU response contains
    an error
    """


class NoMoreRecordsError(SruthiError):
    """
    This error is raised if all records have been loaded (or no records are
    present)
    """


class SruthiWarning(Warning):
    """
    General sruthi warning class to provide a superclass for all warnings
    """


class WrongNamespaceWarning(SruthiWarning):
    """
    A warning to indicate, that a server uses the wrong SRU namespace.
    """



# -*- coding: utf-8 -*-

from . import client


def searchretrieve(url, query, **kwargs):
    search_params = ['query', 'start_record', 'requests_kwargs']
    search_kwargs = {k: v for k, v in kwargs.items() if k in search_params}
    search_kwargs['query'] = query

    # assume all others kwargs are for the client
    client_kwargs = {k: v for k, v in kwargs.items() if k not in search_params}
    client_kwargs['url'] = url

    c = client.Client(**client_kwargs)
    return c.searchretrieve(**search_kwargs)


def explain(url):
    c = client.Client(url)
    return c.explain()
